import logging
from json.decoder import JSONDecodeError
import pandas as pd
import numpy as np
from .EuroLeagueData import EuroLeagueData
from .boxscore_data import BoxScoreData
from .utils import get_requests, get_pbp_lineups

logger = logging.getLogger(__name__)


class PlayByPlay(EuroLeagueData):
    """
    A class for getting the game play-by-play data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_game_play_by_play_data(
        self,
        season: int,
        gamecode: int,
        include_ishometeam: bool = False,
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of a particular game.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

            include_ishometeam (bool, optional): A bool indicator whether to
                include the `IsHomeTeam` column in the returned dataframe,
                which shows whether the action was performed by the home team
                or not. Defaults to False. Introduced for backward
                compatibility.

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of the game.
        """
        url = "https://live.euroleague.net/api/PlaybyPlay"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)
        try:
            data = r.json()
        except JSONDecodeError as exc:
            logger.error(
                f"Game code, {gamecode}, season {season}, "
                "did not return valid JSON data."
            )
            raise exc

        periods = [
            'FirstQuarter', 'SecondQuarter', 'ThirdQuarter', 'ForthQuarter',
            'ExtraTime'
        ]
        all_data = []
        for p, period in enumerate(periods):
            if data[period]:
                df = pd.json_normalize(data[period])
                df["PERIOD"] = p + 1
                all_data.append(df)

        if not all_data:
            logger.warning(
                f"No play-by-play data found for gamecode {gamecode} "
                f" and season {season}"
            )
            return pd.DataFrame()

        pbp_df = pd.concat(all_data).reset_index(drop=True)
        pbp_df["PLAYER"] = (
            pbp_df["PLAYER"].str.replace("  ", " ")
            .str.replace(" , ", ", ").str.strip()
        )
        pbp_df['CODETEAM'] = pbp_df['CODETEAM'].str.strip()
        pbp_df['PLAYER_ID'] = pbp_df['PLAYER_ID'].str.strip()
        pbp_df["PLAYTYPE"] = pbp_df["PLAYTYPE"].str.strip()
        pbp_df["MARKERTIME"] = pbp_df["MARKERTIME"].str.strip()
        if include_ishometeam:  # for backward compatibility
            home_team = data["CodeTeamA"]
            away_team = data["CodeTeamB"]
            pbp_df["IsHomeTeam"] = np.where(
                pbp_df["CODETEAM"] == home_team, True,
                np.where(pbp_df["CODETEAM"] == away_team,
                         False, None)  # type: ignore
            )
        pbp_df.insert(0, 'Season', season)
        pbp_df.insert(1, 'Gamecode', gamecode)
        # insert a TRUE_NUMBEROFPLAY column
        # often the NUMBEROFPLAY column is not in order
        pbp_df["TRUE_NUMBEROFPLAY"] = np.arange(pbp_df.shape[0])

        return pbp_df

    def get_game_play_by_play_data_round(
        self,
        season: int,
        round_number: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a single
        round

        Args:
            season (int): The start year of the season
            round_number (int): The round of the season

        Returns:
            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single round_number
        """
        df = self.get_round_data_from_game_data(
            season, round_number, self.get_game_play_by_play_data)
        return df

    def get_game_play_by_play_data_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a single
        season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_play_by_play_data)
        return data_df

    def get_game_play_by_play_data_range_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a range of
        seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in range of seasons
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_play_by_play_data)
        return df

    def get_game_pbp_data_lineups(
        self,
        season,
        gamecode,
        validate=True
    ) -> pd.DataFrame:
        """
        Get the play-by-play (PBP) data enriched with the teams' lineups for
        every action in the PBP data.

        There are three cases where the player in the corresponding row
        (action) is not part of the lineup:

            1. When the player is subbed "OUT", the assigned lineup does not
            contain player name. Since this is reasonable, the value of the
            `validate_on_court_player` indicator is set to `True`.

            2. A player passes on to a teamate, who draws a shooting foul. If
            the passer is subbed before the first free throw and the first free
            throw is made then the subbed player is given an assist. However,
            the player is not in the extracted lineup because he has already
            been subbed. The value of the `validate_on_court_player`
            indicator is set to `False`. We don't fix the lineup, because it
            breaks the lineup continuinity. It is a quirk of the data
            collection andrecording.

            3. There are a few instances where a sub is recorded many seconds
            and actions since it actually happened. This causes issues, such
            as a player records an actions, such as a rebound, but he comes
            on the court (according to the data) many seconds after. This has
            been validated by watching theactual footage of the game. This
            requires a lot of manual work, which is beyond the score of this
            library. It is another quirk of the datacollection and recording,
            hence the value of the `validate_on_court_player` indicator is set
            to `False`.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

            validate (bool, optional): A bool indicator whether to enrich the
                dataframe with two extra columns, which validate the validity
                and consistency of the extracted lineup. Defaults to True.

        Returns:

            pd.DataFrame: A dataframe with the play-by-play enriched with
                teams' lineups
        """

        # Fetch play-by-play data
        pbp_data = self.get_game_play_by_play_data(
            season=season, gamecode=gamecode, include_ishometeam=True)

        # Get the starting line-ups from boxscore data
        boxscoredata = BoxScoreData(competition=self.competition)
        try:
            game_bxscr_stats = boxscoredata.get_players_boxscore_stats(
                season=season, gamecode=gamecode)
        except Exception as e:  # noqa: E722
            logger.warning(
                f"Something went wrong when fetching boxscore data for "
                f"game {gamecode}, season {season}.\nError message: {e}. "
                "\nSkip and continue"
            )
            game_bxscr_stats = pd.DataFrame()

        pbp_df = get_pbp_lineups(
            pbp_df=pbp_data,
            boxscore_df=game_bxscr_stats,
            validate=validate
        )
        return pbp_df

    def get_game_pbp_data_lineups_round(
        self,
        season: int,
        round_number: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play with lineups data of *all* games
        in a single round

        Args:
            season (int): The start year of the season
            round_number (int): The round of the season

        Returns:
            pd.DataFrame: A dataframe with the play-by-play data with lineups
                of all games in a single round
        """
        df = self.get_round_data_from_game_data(
            season, round_number, self.get_game_pbp_data_lineups)
        return df

    def get_game_pbp_data_lineups_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data enriched with team lineups
        of *all* games in a single season

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in a single season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_game_pbp_data_lineups)
        return data_df

    def get_game_pbp_data_lineups_range_seasons(
        self, start_season: int, end_season: int
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data enriched with team lineups
        of *all* games in a range of seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:

            pd.DataFrame: A dataframe with the play-by-play data of all games
                in range of seasons
        """
        df = self.get_range_seasons_data(
            start_season, end_season, self.get_game_pbp_data_lineups)
        return df



from typing import Optional, List, Callable
import requests
from requests.exceptions import HTTPError
from json.decoder import JSONDecodeError
import logging
import pandas as pd
import numpy as np
from tqdm.auto import tqdm

logging.basicConfig(encoding='utf-8', level=logging.INFO)
logger = logging.getLogger(__name__)


def get_requests(
    url: str,
    params: dict = {},
    headers: dict = {"Accept": "application/json"}
) -> requests.models.Response:
    """
    A wrapper to `requests.get()` which handles unsuccesful requests too.

    Args:

        url (str): _description_

        params (dict, optional): The `params` variables in get requests.
            Defaults to {}.

        headers (dict, optional): the `header` variable in get requests.
            Defaults to {"Accept": "application/json"}.

    Raises:

        Requests Error: If get request was not succesful

    Returns:

        requests.models.Response: The response object.
    """
    r = requests.get(url, params=params, headers=headers, timeout=60)

    if r.status_code != 200:
        r.raise_for_status()

    return r


def raise_error(
    var: Optional[str],
    descripitve_var: str,
    available_vals: List,
    allow_none: bool = False,
) -> None:
    """
    A function that raises a ValueError with specific message.

    Args:

        var (str): The input variable by the user

        descripitve_var (str): A description in plain English of this variable

        available_vals (List): The available variables

        allow_none (bool, optional): If `var` can take None value.
            Defaults to False.

    Raises:

        ValueError: if `var` not applicable
    """

    if allow_none:
        available_vals.append(None)

    if var not in available_vals:
        raise ValueError(
            f"{descripitve_var}, {var}, is not applicable. "
            f"Available values: {available_vals}"
        )
    return


def get_data_over_collection_of_games(
    game_codes_df,
    season: int,
    fun: Callable[[int, int], pd.DataFrame]
) -> pd.DataFrame:
    """A function that collects data over a collection of games given their
    game codes. It is a wrapper function that calls the `fun` function

    Args:
        game_codes_df (pd.DataFrame): A dataframe of the game codes to collect
        season (int, optional): The start year of the season.
        fun (Callable[[int, int], pd.DataFrame]): A callable function that
            determines that type of data to be collected. Available values:
            - get_game_report
            - get_game_stats
            - get_game_teams_comparison
            - get_game_play_by_play_data
            - get_game_shot_data
            - get_game_boxscore_quarter_data
            - get_player_boxscore_stats_data
            - get_game_metadata


    Returns:
        pd.DataFrame: A dataframe with the corresponding data of all
            games in the collection.
    """
    data_list = []
    for _, row in tqdm(game_codes_df.iterrows(),
                       total=game_codes_df.shape[0],
                       desc=f"Season {season}", leave=True):
        game_code = row["gameCode"]
        try:
            df = fun(season, game_code)
            if df.empty:
                logger.warning(
                    f"Game {game_code}, season {season} returned no data."
                )
                continue
            if ("Phase" not in df.columns) and ("Phase" in row):
                df.insert(1, "Phase", row["Phase"])
            if ("Round" not in df.columns) and ("Round" in row):
                df.insert(2, "Round", row["Round"])
            data_list.append(df)
        except HTTPError as err:
            logger.error(
                f"HTTPError: Didn't find gamecode {game_code} for season "
                f"{season}. \nError message {err}. "
                "\nSkip and continue."
            )
        except JSONDecodeError:
            logger.error(
                f"JSONDecodeError: Game code, {game_code}, "
                f"season {season}, did not return valid JSON data. "
                "\nSkip and continue."
            )
        except Exception as e:  # noqa: E722
            logger.error(
                f"\nSomething went wrong for game {game_code}, "
                f"season {season}.\nError message: {e}. "
                "\nSkip and continue"
            )

    if data_list:
        data_df = pd.concat(data_list, axis=0)
        data_df.reset_index(drop=True, inplace=True)
    else:
        data_df = pd.DataFrame([])
    return data_df


def get_pbp_lineups(
    pbp_df: pd.DataFrame,
    boxscore_df: pd.DataFrame,
    validate=True,
) -> pd.DataFrame:
    """
    A function that extracts the lineups from play-by-play data.

    Args:

        pbp_df (pd.DataFrame): The play-by-play dataframe.
        boxscore_df (pd.DataFrame): The boxscore dataframe.
        validate (bool, optional): If to validate if the on-court players

    Returns:

        pd.DataFrame: A dataframe with the lineups.
    """

    def process_sub(five, player, sub_type):
        opp_sub_type = "OUT" if sub_type == "IN" else "IN"
        potential_indx = (
            pbp_df.loc[idx + 1:].index[
                (pbp_df.loc[idx + 1:, "PLAYTYPE"] == opp_sub_type)
            ]
        )
        potential_nops = (
            set(potential_indx).difference(processed_idxs)
        )
        if not potential_nops:
            logger.warning(
                f"No potential matching subs found for gamecode "
                f"{gamecode} and season {season}"
            )
            return five
        matching_idx = min(potential_nops)
        processed_idxs.append(idx)
        processed_idxs.append(matching_idx)
        matching_row = pbp_df.loc[matching_idx]
        invalid_mask = (
            (matching_row["PLAYTYPE"] != opp_sub_type) or
            (matching_row["CODETEAM"] != team) or
            (matching_row["MARKERTIME"] != markertime)
        )
        if invalid_mask:
            logger.warning(
                f"Something went wrong for gamecode {gamecode} at sub "
                f"index {idx} with matching sub index {matching_idx}"
            )
        player_sub = matching_row["PLAYER"]
        player_in = player if sub_type == "IN" else player_sub
        player_out = player_sub if sub_type == "IN" else player
        if player_in == player_out:
            # there are instance where the same player is subbed in and out
            return five
        elif player_out not in five:
            logger.warning(
                f"Player {player_out} not found in current lineup, "
                f"{five}, for gamecode {gamecode} and season {season}."
            )
            return five
        else:
            pindx = five.index(player_out)
            five = five[:pindx] + [player_in] + five[pindx + 1:]
            return five

    def validate_player(x, col1, col2):
        flag = False
        if (x["PLAYER"] is not None) and (x["PLAYTYPE"] != "OUT"):
            if x["PLAYER"] in (x[col1] + x[col2]):
                flag = True
        else:
            flag = True
        return flag

    if pbp_df.empty or (pbp_df is None):
        return pbp_df

    if boxscore_df.empty or (boxscore_df is None):
        pbp_df["validate_on_court_player"] = False
        pbp_df["Lineup_A"] = pbp_df["Lineup_A"].apply(lambda x: [])
        pbp_df["Lineup_B"] = pbp_df["Lineup_B"].apply(lambda x: [])
        return pbp_df

    pbp_df = pbp_df.copy()
    gamecodes = pbp_df["Gamecode"].unique()
    if len(gamecodes) > 1:
        raise ValueError(
            "PBP DataFrame contains multiple game codes. "
            "Please provide data for a single game."
        )
    seasons = pbp_df["Season"].unique()
    if len(seasons) > 1:
        raise ValueError(
            "PBP DataFrame contains multiple seasons. "
            "Please provide data for a single season."
        )
    gamecode = gamecodes[0]
    season = seasons[0]

    # Asign the starting lineups to the first entry of the PBP data.
    pbp_df["Lineup_A"] = None
    pbp_df["Lineup_B"] = None

    # find home and away teams
    hm_aw = boxscore_df[["Home", "Team"]].drop_duplicates()
    home_team = hm_aw.loc[hm_aw["Home"] == 1, "Team"].values[0]
    away_team = hm_aw.loc[hm_aw["Home"] == 0, "Team"].values[0]

    starting_five = boxscore_df.loc[
        boxscore_df["IsStarter"] == 1, ["Team", "Player"]
    ]
    starting_five['ID'] = starting_five.groupby('Team').cumcount()

    # Pivot the DataFrame
    starting_five = starting_five.pivot(
        index='ID', columns='Team', values='Player')

    # Reset index if needed
    starting_five.reset_index(drop=True, inplace=True)
    starting_five = starting_five[[home_team, away_team]]
    # TODO: remove cleaning, this has been done in boxscore data
    starting_five[home_team] = starting_five[home_team].str.replace(
        "  ", " ").str.replace(" , ", ", ").str.strip()
    starting_five[away_team] = starting_five[away_team].str.replace(
        "  ", " ").str.replace(" , ", ", ").str.strip()
    starting_five_dict = starting_five.to_dict(orient='list')

    pbp_df.at[0, "Lineup_A"] = starting_five_dict[home_team]
    pbp_df.at[0, "Lineup_B"] = starting_five_dict[away_team]
    if "IsHomeTeam" not in pbp_df.columns:
        pbp_df["IsHomeTeam"] = np.where(
            pbp_df["CODETEAM"] == home_team, True,
            np.where(pbp_df["CODETEAM"] == away_team,
                     False, None)  # type: ignore
        )
    # start processing the sub entries, row by row.
    processed_idxs: list = []
    current_five_home = starting_five_dict[home_team].copy()
    current_five_away = starting_five_dict[away_team].copy()
    for idx, row in pbp_df.iterrows():
        playtype = row["PLAYTYPE"]
        # there are rare instances of an extra space in player name
        player = (
            row["PLAYER"] if row["PLAYER"] is None
            else row["PLAYER"]
        )
        team = row["CODETEAM"]
        markertime = row["MARKERTIME"]
        if team != "":
            is_home = home_team == team
            five = current_five_home if is_home else current_five_away
            if (playtype == "OUT") and (idx not in processed_idxs):
                five = process_sub(five, player, playtype)
            elif (playtype == "IN") and (idx not in processed_idxs):
                five = process_sub(five, player, playtype)
            if is_home:
                current_five_home = five
            else:
                current_five_away = five

        pbp_df.at[idx, "Lineup_A"] = current_five_home
        pbp_df.at[idx, "Lineup_B"] = current_five_away

    if validate:
        lu_cols = [u for u in pbp_df.columns if u.startswith("Lineup_")]
        pbp_df["validate_on_court_player"] = pbp_df.apply(
            lambda x: validate_player(x, lu_cols[0], lu_cols[1]),
            axis=1
        )

    if "TRUE_NUMBEROFPLAY" in pbp_df.columns:
        pbp_df["TRUE_NUMBEROFPLAY"] = np.arange(pbp_df.shape[0])

    return pbp_df



from typing import Optional
import pandas as pd
from .EuroLeagueData import EuroLeagueData
from .utils import (
    raise_error,
    get_requests
)


class TeamStats(EuroLeagueData):
    """
    A class for getting team-level stats and data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_team_stats(
        self,
        endpoint: str,
        params: dict = {},
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A wrapper functions for collecting teams' stats.
        Allows for three types of data:
        - all seasons
        - single season
        - range of seasons

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            params (Dict[str, Union[str, int]]): A dictionary of the parmaters
                for the get request.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Raises:

            ValueError: If the endpoint is not applicable

            ValueError: If the phase_type_code is not applicable

            ValueError: If the statistic_mode is not applicable

        Returns:

            pd.DataFrame: A dataframe with the teams' stats.
        """

        available_endpoints = [
            "traditional", "advanced",
            "opponentsTraditional", "opponentsAdvanced"
        ]
        available_phase_type_code = ["RS", "PO", "FF"]
        available_stat_mode = ["PerGame", "Accumulated"]

        raise_error(
            endpoint, "Statistic type", available_endpoints, False)
        raise_error(
            statistic_mode, "Statistic Aggregation", available_stat_mode,
            False)
        raise_error(
            phase_type_code, "Phase type code", available_phase_type_code,
            True)

        params["statisticMode"] = statistic_mode
        params["phaseTypeCode"] = phase_type_code
        params["limit"] = 400

        url_ = f"{self.url}/statistics/teams/{endpoint}"

        r = get_requests(url_, params=params)
        data = r.json()
        if data["total"] < len(data["teams"]):
            params["limit"] = len(data["teams"]) + 1
            r = get_requests(url_, params=params)
            data = r.json()
        df = pd.json_normalize(data["teams"])
        return df

    def get_team_stats_all_seasons(
        self,
        endpoint: str,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A function that gets the team stats for all seasons

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the teams' stats
        """
        params = {"SeasonMode": "All"}
        df = self.get_team_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_team_stats_single_season(
        self,
        endpoint: str,
        season: int,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A function that returns the teams' stats in a single season

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            season (int): The start year of the season

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the teams' stats
        """
        params = {
            "SeasonMode": "Single",
            "SeasonCode": f"{self.competition}{season}",
        }
        df = self.get_team_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_team_stats_range_seasons(
        self,
        endpoint: str,
        start_season: int,
        end_season: int,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame"
    ) -> pd.DataFrame:
        """
        A function that returns the teams' stats in a range of seasons

        Args:

            endpoint (str): The type of stats to fetch. Available values:
                - traditional
                - advanced
                - opponentsTraditional
                - opponentsAdvanced

            start_season (int): The start year of the start season

            end_season (int): The end year of the end season

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                Defaults to "PerGame".

        Returns:

            pd.DataFrame: A dataframe with the teams' stats
        """
        params = {
            "SeasonMode": "Range",
            "FromSeasonCode": f"{self.competition}{start_season}",
            "ToSeasonCode": f"{self.competition}{end_season}",
        }
        df = self.get_team_stats(
            endpoint, params, phase_type_code, statistic_mode)
        return df

    def get_team_stats_leaders(
        self,
        params: dict = {},
        stat_category: str | None = None,
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None
    ) -> pd.DataFrame:
        """
        A wrapper function for collecting the team leaders in a given
        stat category.

        We are using the v2 endpoint
        `/v2​/competitions​/{competitionCode}​/stats​/clubs​/leaders`
        instead of the v3 endpoint
        `/v3​/competitions​/{competitionCode}​/statistics​/teams​/leaders`
        because it's richer in options, data slices and statistical
        metrics.


        Args:

            params (Dict[str, Union[str, int]]): A dictionary of parameters
                for the get request.

            top_n (int): The number of top N teams to return.
                Defaults to 200.

            stat_category (str): The stat category. Available values:
                - None # (time played)
                - Valuation
                - Score
                - TotalRebounds
                - OffensiveRebounds
                - DefensiveRebounds
                - Assistances
                - Steals
                - BlocksFavour
                - BlocksAgainst
                - Turnovers
                - FoulsReceived
                - FoulsCommited
                - FreeThrowsMade
                - FreeThrowsAttempted
                - FreeThrowsPercent
                - FieldGoalsMade2
                - FieldGoalsAttempted2
                - FieldGoals2Percent
                - FieldGoalsMade3
                - FieldGoalsAttempted3
                - FieldGoals3Percent
                - FieldGoalsMadeTotal
                - FieldGoalsAttemptedTotal
                - FieldGoalsPercent
                - AccuracyMade
                - AccuracyAttempted
                - AccuracyPercent
                - AssitancesTurnoversRation
                - GamesPlayed
                - GamesStarted
                - TimePlayed
                - Contras
                - Dunks
                - OffensiveReboundPercentage
                - DefensiveReboundPercentage
                - ReboundPercentage
                - EffectiveFeildGoalPercentage
                - TrueShootingPercentage
                - AssistRatio
                - TurnoverRatio
                - FieldGoals2AttemptedRatio
                - FieldGoals3AttemptedRatio
                - FreeThrowRate
                - Possessions
                - GamesWon
                - GamesLost
                - DoubleDoubles
                - TripleDoubles
                - FieldGoalsAttempted2Share
                - FieldGoalsAttempted3Share
                - FreeThrowsAttemptedShare
                - FieldGoalsMade2Share
                - FieldGoalsMade3Share
                - FreeThrowsMadeShare
                - PointsMade2Rate
                - PointsMade3Rate
                - PointsMadeFreeThrowsRate
                - PointsAttempted2Rate
                - PointsAttempted3Rate
                - Age

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - PerMinute
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

        Raises:

            ValueError: If the stat_category is not applicable

            ValueError: If the phase_type_code is not applicable

            ValueError: If the statistic_mode is not applicable

            ValueError: If the game_type is not applicable

        Returns:

            pd.DataFrame: A dataframe with the top teams' stats
        """
        avaiable_stat_category = [
            None,  # (time played)
            "Valuation",
            "Score",
            "TotalRebounds",
            "OffensiveRebounds",
            "DefensiveRebounds",
            "Assistances",
            "Steals",
            "BlocksFavour",
            "BlocksAgainst",
            "Turnovers",
            "FoulsReceived",
            "FoulsCommited",
            "FreeThrowsMade",
            "FreeThrowsAttempted",
            "FreeThrowsPercent",
            "FieldGoalsMade2",
            "FieldGoalsAttempted2",
            "FieldGoals2Percent",
            "FieldGoalsMade3",
            "FieldGoalsAttempted3",
            "FieldGoals3Percent",
            "FieldGoalsMadeTotal",
            "FieldGoalsAttemptedTotal",
            "FieldGoalsPercent",
            "AccuracyMade",
            "AccuracyAttempted",
            "AccuracyPercent",
            "AssitancesTurnoversRation",
            "GamesPlayed",
            "GamesStarted",
            "TimePlayed",
            "Contras",
            "Dunks",
            "OffensiveReboundPercentage",
            "DefensiveReboundPercentage",
            "ReboundPercentage",
            "EffectiveFeildGoalPercentage",
            "TrueShootingPercentage",
            "AssistRatio",
            "TurnoverRatio",
            "FieldGoals2AttemptedRatio",
            "FieldGoals3AttemptedRatio",
            "FreeThrowRate",
            "Possessions",
            "GamesWon",
            "GamesLost",
            "DoubleDoubles",
            "TripleDoubles",
            "FieldGoalsAttempted2Share",
            "FieldGoalsAttempted3Share",
            "FreeThrowsAttemptedShare",
            "FieldGoalsMade2Share",
            "FieldGoalsMade3Share",
            "FreeThrowsMadeShare",
            "PointsMade2Rate",
            "PointsMade3Rate",
            "PointsMadeFreeThrowsRate",
            "PointsAttempted2Rate",
            "PointsAttempted3Rate",
            "Age"
        ]
        available_phase_type_code = ["RS", "PO", "FF"]
        available_stat_mode = [
            "PerGame",
            "Accumulated",
            "PerMinute",
            "Per100Possesions",
            "PerGameReverse",
            "AccumulatedReverse"
        ]
        available_game_types = [
            "HomeGames",
            "AwayGames",
            "GamesWon",
            "GamesLost",
        ]

        raise_error(stat_category, "Stat category",
                    avaiable_stat_category, False)
        raise_error(
            statistic_mode, "Statistic Aggregation", available_stat_mode,
            False)
        raise_error(
            phase_type_code, "Phase type code", available_phase_type_code,
            True)
        raise_error(game_type, "Game type", available_game_types, True)

        params["category"] = stat_category
        params["phaseTypeCode"] = phase_type_code
        params["statisticMode"] = statistic_mode
        params["limit"] = top_n

        url_ = f"{self.url_v2}/stats/clubs/leaders"

        r = get_requests(url_, params=params)
        data = r.json()
        df = pd.json_normalize(data["data"])
        return df

    def get_team_stats_leaders_all_seasons(
        self,
        stat_category: str | None = None,
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Get the top leaders in a statistical category in all seasons

        Args:

            stat_category (str): The stat category. See function
                `self.get_team_stats_leaders` for a list of available stats.

            top_n (int): The number of top N teams to return.
                Defaults to 200.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - PerMinute
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

        Returns:

            pd.DataFrame: A dataframe with the top leading teams and their
                stat
        """
        params = {"SeasonMode": "All"}
        df = self.get_team_stats_leaders(
            params,
            stat_category,
            top_n,
            phase_type_code,
            statistic_mode,
            game_type
        )
        return df

    def get_team_stats_leaders_single_season(
        self,
        season: int,
        stat_category: str = "Score",
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Get the top leaders in a statistical category in a single season

        Args:

            season (int): The start year of the season.

            stat_category (str): The stat category. See function
                `self.get_team_stats_leaders` for a list of available stats.
            top_n (int): The number of top N teams to return.
                Defaults to 200.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - PerMinute
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

        Returns:

            pd.DataFrame: A dataframe with the top leading teams and their
                stat
        """
        params = {
            "SeasonMode": "Single",
            "SeasonCode": f"{self.competition}{season}",
        }
        df = self.get_team_stats_leaders(
            params,
            stat_category,
            top_n,
            phase_type_code,
            statistic_mode,
            game_type
        )
        return df

    def get_team_stats_leaders_range_seasons(
        self,
        start_season: int,
        end_season: int,
        stat_category: str = "Score",
        top_n: int = 200,
        phase_type_code: Optional[str] = None,
        statistic_mode: str = "PerGame",
        game_type: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Get the top leaders in a statistical category in a range of seasons

        Args:

            start_season (int): The start year of the first season in the
                range.

            end_season (int): The end year of the last season in the range.

            stat_category (str): The stat category. See function
                `self.get_team_stats_leaders` for a list of available stats.

            top_n (int): The number of top N teams to return.
                Defaults to 200.

            phase_type_code (Optional[str], optional): The phase of the season,
                available variables:
                - "RS" (regular season)
                - "PO" (play-off)
                - "FF" (final four)
                Defaults to None, which includes all phases.

            statistic_mode (str, optional): The aggregation of statistics,
                available variables:
                - PerGame
                - Accumulated
                - PerMinute
                - Per100Possesions
                - PerGameReverse
                - AccumulatedReverse
                Defaults to "PerGame".

            game_type (Optional[str], optional): The type of games to draw the
                top stats from. Available values:
                - HomeGames
                - AwayGames
                - GamesWon
                - GamesLost
                Defaults to None, meaning all games

        Returns:

            pd.DataFrame: A dataframe with the top leading teams and their
                stat
        """
        params = {
            "SeasonMode": "Range",
            "FromSeasonCode": f"{self.competition}{start_season}",
            "ToSeasonCode": f"{self.competition}{end_season}",
        }
        df = self.get_team_stats_leaders(
            params,
            stat_category,
            top_n,
            phase_type_code,
            statistic_mode,
            game_type
        )
        return df



import pandas as pd
import xmltodict
from .EuroLeagueData import EuroLeagueData
from .utils import get_requests


class Schedule(EuroLeagueData):
    """
    A class for getting schedule data.

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_schedule(self, season: int) -> pd.DataFrame:
        """
        Get the schedule of a season. The schedule includes the date,
        time, and the teams playing in each game.

        The schedule gets updated as the season progresses.

        Args:

            season (int): The start year of the season

        Returns:

            pd.DataFrame: A dataframe with the schedule of the games
        """
        url = f"{self.BASE_URL}/{self.V1}/schedules"
        params = {
            "seasonCode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)
        schedule_dict = xmltodict.parse(r.text)
        df = pd.DataFrame(schedule_dict["schedule"]["item"])
        df["gameday"] = df["gameday"].astype(int)
        return df



from typing import List
import logging
from json.decoder import JSONDecodeError
import pandas as pd
import numpy as np
from .EuroLeagueData import EuroLeagueData
from .utils import (
    get_requests,
    raise_error
)

logger = logging.getLogger(__name__)


class BoxScoreData(EuroLeagueData):
    """
    A class for getting box-score data

    Args:
        competition (str, optional): The competition code, inherited from the
            `EuroLeagueData` class. Choose one of:
            - 'E' for Euroleague
            - 'U' for Eurocup
            Defaults to "E".
    """

    def get_boxscore_data(
        self,
        season: int,
        gamecode: int,
        boxscore_type: str = "ByQuarter"
    ) -> List[dict]:
        """A helper function that gets the boxscore data of a particular data.

        Args:
            season (int): The start year of the season
            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.
            boxscore_type (str, optional): The type of quarter boxscore data.
                Available values:
                - Stats
                - ByQuarter
                - EndOfQuarter
                Defaults to "ByQuarter".

        Raises:
            ValueError: If boxscore_type value is not valid.

        Returns:
            List[dict]: A list of dictionaries with the data.
        """
        url = "https://live.euroleague.net/api/Boxscore"
        params = {
            "gamecode": gamecode,
            "seasoncode": f"{self.competition}{season}"
        }
        r = get_requests(url, params=params)

        try:
            data = r.json()
        except JSONDecodeError as exc:
            logger.error(
                f"Game code, {gamecode}, season {season}, "
                "did not return valid JSON data."
            )
            raise exc
        except KeyError as exc:
            logger.error(
                f"Game code, {gamecode}, season {season}, "
                "returned incomplete data."
            )
            raise exc

        boxscore_types = ["Stats", "ByQuarter", "EndOfQuarter"]

        if boxscore_type not in boxscore_types:
            raise_error(boxscore_type, "Boxscore", boxscore_types, False)

        return data[boxscore_type]

    def get_teams_boxscore_quarter_scores(
        self,
        season: int,
        gamecode: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the boxscore quarterly data of a particular game.

        Args:

            season (int): The start year of the season

            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

            boxscore_type (str): The type of quarter boxscore data.
                Available values:
                - ByQuarter
                - EndOfQuarter
                Default: ByQuarter

        Raises:
            ValueError: If boxscore_type value is not valid.

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of the
                game.
        """
        valid_vals = ["ByQuarter", "EndOfQuarter"]
        if boxscore_type not in valid_vals:
            raise_error(boxscore_type, "Boxscore quarter type",
                        valid_vals, False)

        data = self.get_boxscore_data(season, gamecode, boxscore_type)
        df = pd.json_normalize(data)
        df.insert(0, 'Season', season)
        df.insert(1, 'Gamecode', gamecode)
        return df

    def get_players_boxscore_stats(
        self,
        season: int,
        gamecode: int
    ) -> pd.DataFrame:
        """
        The players' and team's total stats of a particular game.

        Args:
            season (int): The start year of the season
            gamecode (int): The game-code of the game of interest.
                It can be found on Euroleague's website.

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats
        """

        def dict_to_df_bx(datadict, home=1):
            playerstats_df = pd.json_normalize(datadict["PlayersStats"])
            teamstats_df = pd.json_normalize(datadict["tmr"])
            totalstats_df = pd.json_normalize(datadict["totr"])
            df = pd.concat(
                [
                    # fix types to avoid pandas futurewarning
                    playerstats_df.astype(teamstats_df.dtypes),
                    teamstats_df,
                    totalstats_df
                ],
                ignore_index=True
            )
            if "Plusminus" in df.columns:
                # replace None entries and fix data type
                # this behaviour used to be automatic, but since version
                # 2.1.0, pandas throws a futurewarning
                df.loc[df["Plusminus"].isnull(), "Plusminus"] = np.nan
                df["Plusminus"] = df["Plusminus"].astype(float)
            df.iloc[-2:, 0] = ["Team", "Total"]  # type: ignore
            df.iloc[-2:, 5] = ["Team", "Total"]  # type: ignore
            df["Team"] = df["Team"].ffill()
            df.insert(0, 'Season', season)
            df.insert(1, 'Gamecode', gamecode)
            df.insert(2, "Home", home)
            return df

        data = self.get_boxscore_data(season, gamecode, "Stats")

        home_df = dict_to_df_bx(data[0], home=1)
        away_df = dict_to_df_bx(data[1], home=0)

        df = pd.concat([home_df, away_df], axis=0, ignore_index=True)
        df["Player"] = (
            df["Player"].str.replace("  ", " ")
            .str.replace(" , ", ", ").str.strip()
        )
        return df

    def get_teams_boxscore_quarter_scores_round(
        self,
        season: int,
        round_number: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the boxscore quarter data of all games in a
        particular round of a season

        Args:

            season (int): The start year of the season

            round_number (int): The number of the round

            boxscore_type (str): The type of quarter boxscore data.
                Available values:
                - ByQuarter
                - EndOfQuarter
                Default: ByQuarter

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of all
                games in a particular round of a season
        """
        get_teams_boxscore_quarter_scores_ = (
            lambda season, gamecode: self.get_teams_boxscore_quarter_scores(
                season, gamecode, boxscore_type)
        )
        data_df = self.get_round_data_from_game_data(
            season, round_number, get_teams_boxscore_quarter_scores_)
        return data_df

    def get_teams_boxscore_quarter_scores_single_season(
        self,
        season: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the boxscore quarter data of *all* games in a
        single season

        Args:

            season (int): The start year of the season

            boxscore_type (str): The type of quarter boxscore data.
            Available values:
                - ByQuarter
                - EndOfQuarter
            Default: ByQuarter

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of all
                games in a single season
        """
        get_teams_boxscore_quarter_scores_ = (
            lambda season, gamecode: self.get_teams_boxscore_quarter_scores(
                season, gamecode, boxscore_type)
        )
        data_df = self.get_season_data_from_game_data(
            season, get_teams_boxscore_quarter_scores_)
        return data_df

    def get_teams_boxscore_quarter_scores_range_seasons(
        self,
        start_season: int,
        end_season: int,
        boxscore_type: str = "ByQuarter"
    ) -> pd.DataFrame:
        """
        A function that gets the play-by-play data of *all* games in a range of
        seasons

        Args:

            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

            boxscore_type (str): The type of quarter boxscore data.
                Available values:
                - ByQuarter
                - EndOfQuarter
                Default: ByQuarter

        Returns:

            pd.DataFrame: A dataframe with the boxscore quarter data of all
                games in range of seasons
        """
        get_teams_boxscore_quarter_scores_ = (
            lambda season, gamecode: self.get_teams_boxscore_quarter_scores(
                season, gamecode, boxscore_type)
        )
        df = self.get_range_seasons_data(
            start_season, end_season, get_teams_boxscore_quarter_scores_)
        return df

    def get_players_boxscore_stats_round(
        self,
        season: int,
        round_number: int
    ) -> pd.DataFrame:
        """
        A function that return the player boxscore stats for all games in a
        single round

        Args:
            season (int): The start year of the start season
            round_number (int): The number of the round

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats for
                a season
        """
        data_df = self.get_round_data_from_game_data(
            season, round_number, self.get_players_boxscore_stats)
        return data_df

    def get_players_boxscore_stats_single_season(
        self,
        season: int
    ) -> pd.DataFrame:
        """
        A function that return the player boxscore stats for all games in a
        single season

        Args:
            season (int): The start year of the start season

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats for
                a season
        """
        data_df = self.get_season_data_from_game_data(
            season, self.get_players_boxscore_stats)
        return data_df

    def get_players_boxscore_stats_range_seasons(
        self,
        start_season: int,
        end_season: int
    ) -> pd.DataFrame:
        """
        A function that return the player boxscore stats for all games in
        multiple season

        Args:
            start_season (int): The start year of the start season

            end_season (int): The start year of the end season

        Returns:
            pd.DataFrame: A dataframe with home and away team player stats for
                a season
        """
        data_df = self.get_range_seasons_data(
            start_season, end_season, self.get_players_boxscore_stats)
        return data_df
